//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include <vector>
#include <set>

#include "parallel/PetscSolver.h"
#include "parallel/StdMpi.h"
#include "parallel/ParallelDebug.h"
#include "DOFVector.h"
#include "Debug.h"
#include "SystemVector.h"

#include "petscksp.h"

namespace AMDiS {

  PetscErrorCode myKSPMonitor(KSP ksp, PetscInt iter, PetscReal rnorm, void *)
  {    
    if (iter % 100 == 0 && MPI::COMM_WORLD.Get_rank() == 0)
      std::cout << "[0]  Petsc-Iteration " << iter << ": " << rnorm << std::endl;

    return 0;
  }
 

  void PetscSolver::solve(AdaptInfo *adaptInfo, bool fixedMatrix)
  {
    FUNCNAME("PetscSolver::solve()");

    TEST_EXIT(meshDistributor)("Should not happen!\n");

    double wtime = MPI::Wtime();

    fillPetscMatrix(systemMatrix, rhs);
    solvePetscMatrix(*solution, adaptInfo);   

    INFO(info, 8)("solution of discrete system needed %.5f seconds\n", 
		  MPI::Wtime() - wtime);
  }


  void PetscSolver::setDofMatrix(DOFMatrix* mat, int dispMult, 
				 int dispAddRow, int dispAddCol)
  {
    FUNCNAME("PetscSolver::setDofMatrix()");

    TEST_EXIT(mat)("No DOFMatrix!\n");

    typedef map<DegreeOfFreedom, DegreeOfFreedom> DofMapping;
    typedef map<BoundaryType, DofMapping> PeriodicDofMap;

    StdMpi<PeriodicDofMap> stdMpi(meshDistributor->getMpiComm());
    if (meshDistributor->getMpiRank() == 0) {
      for (int i = 1; i < meshDistributor->getMpiSize(); i++)
	stdMpi.recv(i);
    } else {
      stdMpi.send(0, meshDistributor->getPeriodicMapping());
    }
    stdMpi.startCommunication<int>(MPI_INT);



    StdMpi<PeriodicDofMap> stdMpi2(meshDistributor->getMpiComm());
    PeriodicDofMap overallDofMap;

    if (meshDistributor->getMpiRank() == 0) {
      overallDofMap = meshDistributor->getPeriodicMapping();
      for (int i = 1; i < meshDistributor->getMpiSize(); i++) {
	PeriodicDofMap &rankDofMap = stdMpi.getRecvData(i);
	for (PeriodicDofMap::iterator it = rankDofMap.begin(); it != rankDofMap.end(); ++it) {
	  for (DofMapping::iterator dofIt = it->second.begin(); dofIt != it->second.end(); ++dofIt) {
	    if (overallDofMap[it->first].count(dofIt->first) == 1) {
	      TEST_EXIT_DBG(overallDofMap[it->first][dofIt->first] == dofIt->second)
		("Should not happen!\n");
	    } else {
	      overallDofMap[it->first][dofIt->first] = dofIt->second;
	    }
	  }
	}
      }

      for (int i = 1; i < meshDistributor->getMpiSize(); i++)
	stdMpi2.send(i, overallDofMap);
    } else {
      stdMpi2.recv(0);
    }


    stdMpi2.startCommunication<int>(MPI_INT);
    
    if (meshDistributor->getMpiRank() > 0)
      overallDofMap = stdMpi2.getRecvData(0);
   

    using mtl::tag::row; using mtl::tag::nz; using mtl::begin; using mtl::end;
    namespace traits= mtl::traits;
    typedef DOFMatrix::base_matrix_type Matrix;

    traits::col<Matrix>::type col(mat->getBaseMatrix());
    traits::const_value<Matrix>::type value(mat->getBaseMatrix());

    typedef traits::range_generator<row, Matrix>::type cursor_type;
    typedef traits::range_generator<nz, cursor_type>::type icursor_type;

    std::vector<int> cols;
    std::vector<double> values;
    cols.reserve(300);
    values.reserve(300);
    
    std::vector<int> globalCols;

    // === Traverse all rows of the dof matrix and insert row wise the values ===
    // === to the petsc matrix.                                               ===

    for (cursor_type cursor = begin<row>(mat->getBaseMatrix()), 
	   cend = end<row>(mat->getBaseMatrix()); cursor != cend; ++cursor) {

      // Global index of the current row dof.
      int globalRowDof = meshDistributor->mapLocalToGlobal(*cursor);
      // Test if the current row dof is a periodic dof.
      bool periodicRow = meshDistributor->isPeriodicDof(globalRowDof);

      if (!periodicRow) {
	// Calculate petsc row index.
	int rowIndex = globalRowDof * dispMult + dispAddRow;

	for (icursor_type icursor = begin<nz>(cursor), icend = end<nz>(cursor); 
	     icursor != icend; ++icursor) {

	  // Global index of the current column index.
	  int globalColDof = meshDistributor->mapLocalToGlobal(col(*icursor));
	  // Test if the current col dof is a periodic dof.
	  bool periodicCol = meshDistributor->isPeriodicDof(globalColDof);

	  if (!periodicCol) {
	    // Calculate the exact position of the column index in the petsc matrix.
	    int colIndex = globalColDof * dispMult + dispAddCol;
	    MatSetValue(petscMatrix, rowIndex, colIndex, value(*icursor), ADD_VALUES);
	  } else {
	    std::set<int>& perColAsc = meshDistributor->getPerDofAssociations(globalColDof);
	    std::set<int> perAsc;
	    for (std::set<int>::iterator it = perColAsc.begin(); it != perColAsc.end(); ++it)
	      if (*it >= -3)
		perAsc.insert(*it);
	    
	    double scaledValue = value(*icursor) * std::pow(0.5, static_cast<double>(perAsc.size()));
	    std::vector<int> newCols;
	    newCols.push_back(globalColDof);
	    
	    for (std::set<int>::iterator it = perAsc.begin(); it != perAsc.end(); ++it) {
	      int nCols = static_cast<int>(newCols.size());
	      for (int i = 0; i < nCols; i++) {
		TEST_EXIT(overallDofMap[*it].count(newCols[i]) > 0)("Should not happen: %d %d\n", *it, newCols[i]);

		newCols.push_back(overallDofMap[*it][newCols[i]]);
	      }
	    }

	    for (int i = 0; i < newCols.size(); i++) {
	      int colIndex = newCols[i] * dispMult + dispAddCol;
	      MatSetValue(petscMatrix, rowIndex, colIndex, scaledValue, ADD_VALUES);
	    }
	  }
	}
      } else {
	for (icursor_type icursor = begin<nz>(cursor), icend = end<nz>(cursor); 
	     icursor != icend; ++icursor) {	  

	  // Global index of the current column index.
	  int globalColDof = meshDistributor->mapLocalToGlobal(col(*icursor));

	  std::set<int>& perColAsc = meshDistributor->getPerDofAssociations(globalColDof);
	  std::set<int> perAsc;
	  for (std::set<int>::iterator it = perColAsc.begin(); it != perColAsc.end(); ++it)
	    if (*it >= -3)
	      perAsc.insert(*it);

	  std::set<int>& perRowAsc = meshDistributor->getPerDofAssociations(globalRowDof);
	  for (std::set<int>::iterator it = perRowAsc.begin(); it != perRowAsc.end(); ++it)
	    if (*it >= -3)
	      perAsc.insert(*it);

	  double scaledValue = value(*icursor) * std::pow(0.5, static_cast<double>(perAsc.size()));
	  std::vector<std::pair<int, int> > entry;
	  entry.push_back(std::make_pair(globalRowDof, globalColDof));

	  for (std::set<int>::iterator it = perAsc.begin(); it != perAsc.end(); ++it) {
	    int nEntry = static_cast<int>(entry.size());
	    for (int i = 0; i < nEntry; i++) {
	      int perRowDof = 0;
	      if (overallDofMap[*it].count(entry[i].first) > 0)
		perRowDof = overallDofMap[*it][entry[i].first];
	      else
		perRowDof = entry[i].first;

	      int perColDof;
	      if (overallDofMap[*it].count(entry[i].second) > 0)
		perColDof = overallDofMap[*it][entry[i].second];
	      else
		perColDof = entry[i].second;	      	      

	      entry.push_back(std::make_pair(perRowDof, perColDof));
	    }
	  }

	  for (std::vector<std::pair<int, int> >::iterator eIt = entry.begin(); 
	       eIt != entry.end(); ++eIt) {
	    // Calculate petsc row index.
	    int rowIndex = eIt->first * dispMult + dispAddRow;
	    int colIndex = eIt->second * dispMult + dispAddCol;
	    MatSetValue(petscMatrix, rowIndex, colIndex, scaledValue, ADD_VALUES);
	  }
	}
      }
    }
  }


  void PetscSolver::setDofVector(Vec& petscVec, DOFVector<double>* vec, 
				 int dispMult, int dispAdd)
  {
    FUNCNAME("PetscSolver::setDofVector()");

    // Traverse all used dofs in the dof vector.
    DOFVector<double>::Iterator dofIt(vec, USED_DOFS);
    for (dofIt.reset(); !dofIt.end(); ++dofIt) {
      // Calculate global row index of the dof.
      DegreeOfFreedom globalRowDof = 
	meshDistributor->mapLocalToGlobal(dofIt.getDOFIndex());
      // Calculate petsc index of the row dof.
      int index = globalRowDof * dispMult + dispAdd;

      if (meshDistributor->isPeriodicDof(globalRowDof)) {
	std::set<int>& perAsc = meshDistributor->getPerDofAssociations(globalRowDof);
	double value = *dofIt / (perAsc.size() + 1.0);
	VecSetValues(petscVec, 1, &index, &value, ADD_VALUES);

	for (std::set<int>::iterator perIt = perAsc.begin(); perIt != perAsc.end(); ++perIt) {
	  int mappedDof = meshDistributor->getPeriodicMapping(*perIt, globalRowDof);
	  int mappedIndex = mappedDof * dispMult + dispAdd;
	  VecSetValues(petscVec, 1, &mappedIndex, &value, ADD_VALUES);
	}
      } else {
	// The dof index is not periodic.
	double value = *dofIt;
	VecSetValues(petscVec, 1, &index, &value, ADD_VALUES);
      }
    }
  }


  void PetscSolver::createPetscNnzStructure(Matrix<DOFMatrix*> *mat)
  {
    FUNCNAME("PetscSolver::createPetscNnzStructure()");

    TEST_EXIT_DBG(!d_nnz)("There is something wrong!\n");
    TEST_EXIT_DBG(!o_nnz)("There is something wrong!\n");

    int nRankRows = meshDistributor->getNumberRankDofs() * nComponents;
    d_nnz = new int[nRankRows];
    o_nnz = new int[nRankRows];
    for (int i = 0; i < nRankRows; i++) {
      d_nnz[i] = 0;
      o_nnz[i] = 0;
    }

    using mtl::tag::row; using mtl::tag::nz; using mtl::begin; using mtl::end;
    namespace traits = mtl::traits;
    typedef DOFMatrix::base_matrix_type Matrix;
    typedef std::vector<std::pair<int, int> > MatrixNnzEntry;

    // Stores to each rank a list of nnz entries (i.e. pairs of row and column index)
    // that this rank will send to. This nnz entries will be assembled on this rank,
    // but because the row DOFs are not DOFs of this rank they will be send to the
    // owner of the row DOFs.
    std::map<int, MatrixNnzEntry> sendMatrixEntry;


    // First, create for all ranks we send data to MatrixNnzEntry object with 0 entries.
    typedef std::map<int, DofContainer> RankToDofContainer;
    RankToDofContainer& recvDofs = meshDistributor->getRecvDofs();
    for (RankToDofContainer::iterator it = recvDofs.begin();
	 it != recvDofs.end(); ++it)
      sendMatrixEntry[it->first].resize(0);


    for (int i = 0; i < nComponents; i++) {
      for (int j = 0; j < nComponents; j++) {
 	if ((*mat)[i][j]) {
	  Matrix bmat = (*mat)[i][j]->getBaseMatrix();

	  traits::col<Matrix>::type col(bmat);
	  traits::const_value<Matrix>::type value(bmat);
	  
	  typedef traits::range_generator<row, Matrix>::type cursor_type;
	  typedef traits::range_generator<nz, cursor_type>::type icursor_type;
	  
	  for (cursor_type cursor = begin<row>(bmat), 
		 cend = end<row>(bmat); cursor != cend; ++cursor) {

	    // Map the local row number to the global DOF index and create from it
	    // the global PETSc row index of this DOF.
	    int petscRowIdx = 
	      meshDistributor->mapLocalToGlobal(*cursor) * nComponents + i;

	    if (meshDistributor->getIsRankDof(*cursor)) {

	      // === The current row DOF is a rank dof, so create the corresponding ===
	      // === nnz values directly on rank's nnz data.                        ===

	      // This is the local row index of the local PETSc matrix.
	      int localPetscRowIdx = 
		petscRowIdx - meshDistributor->getRstart() * nComponents;

	      TEST_EXIT_DBG(localPetscRowIdx >= 0 && localPetscRowIdx < nRankRows)
		("Should not happen! \n Debug info: localRowIdx = %d   globalRowIndx = %d   petscRowIdx = %d   localPetscRowIdx = %d   rStart = %d   nCompontens = %d   nRankRows = %d\n",
		 *cursor, meshDistributor->mapLocalToGlobal(*cursor), petscRowIdx, localPetscRowIdx, meshDistributor->getRstart(), nComponents, nRankRows);
	      
	      // Traverse all non zero entries in this row.
	      for (icursor_type icursor = begin<nz>(cursor), 
		     icend = end<nz>(cursor); icursor != icend; ++icursor) {
		int petscColIdx = 
		  meshDistributor->mapLocalToGlobal(col(*icursor)) * nComponents + j;

		if (value(*icursor) != 0.0 || petscRowIdx == petscColIdx) {
		  // The row DOF is a rank DOF, if also the column is a rank DOF, 
		  // increment the d_nnz values for this row, otherwise the o_nnz value.
		  if (petscColIdx >= meshDistributor->getRstart() * nComponents && 
		      petscColIdx < meshDistributor->getRstart() * nComponents + nRankRows)
		    d_nnz[localPetscRowIdx]++;
		  else
		    o_nnz[localPetscRowIdx]++;
		}    
	      }
	    } else {
	      typedef std::map<int, DofContainer> RankToDofContainer;

	      // === The current row DOF is not a rank dof, i.e., it will be created ===
	      // === on this rank, but after this it will be send to another rank    ===
	      // === matrix. So we need to send also the corresponding nnz structure ===
	      // === of this row to the corresponding rank.                          ===

	      // Find out who is the member of this DOF.
	      int sendToRank = -1;
	      for (RankToDofContainer::iterator it = meshDistributor->getRecvDofs().begin();
		   it != meshDistributor->getRecvDofs().end(); ++it) {
		for (DofContainer::iterator dofIt = it->second.begin();
		     dofIt != it->second.end(); ++dofIt) {
		  if (**dofIt == *cursor) {
		    sendToRank = it->first;
		    break;
		  }
		}

		if (sendToRank != -1)
		  break;
	      }

	      TEST_EXIT_DBG(sendToRank != -1)("Should not happen!\n");

	      // Send all non zero entries to the member of the row DOF.
	      for (icursor_type icursor = begin<nz>(cursor), 
		     icend = end<nz>(cursor); icursor != icend; ++icursor) {
		if (value(*icursor) != 0.0) {
		  int petscColIdx = 
		    meshDistributor->mapLocalToGlobal(col(*icursor)) * nComponents + j;
		  
		  sendMatrixEntry[sendToRank].
		    push_back(std::make_pair(petscRowIdx, petscColIdx));
		}
	      }

	    } // if (isRankDof[*cursor]) ... else ...
	  } // for each row in mat[i][j]
	} // if mat[i][j]
      } 
    }

    // === Send and recv the nnz row structure to/from other ranks. ===

    StdMpi<MatrixNnzEntry> stdMpi(meshDistributor->getMpiComm(), true);
    stdMpi.send(sendMatrixEntry);
    stdMpi.recv(meshDistributor->getSendDofs());
    stdMpi.startCommunication<int>(MPI_INT);


    // === Evaluate the nnz structure this rank got from other ranks and add it to ===
    // === the PETSc nnz data structure.                                           ===

    for (std::map<int, MatrixNnzEntry>::iterator it = stdMpi.getRecvData().begin();
	 it != stdMpi.getRecvData().end(); ++it) {
      if (it->second.size() > 0) {
	for (unsigned int i = 0; i < it->second.size(); i++) {
	  int r = it->second[i].first;
	  int c = it->second[i].second;

	  int localRowIdx = r - meshDistributor->getRstart() * nComponents;

	  TEST_EXIT_DBG(localRowIdx >= 0 && localRowIdx < nRankRows)
	    ("Got row index %d/%d (nRankRows = %d) from rank %d. Should not happen!\n",
	     r, localRowIdx, nRankRows, it->first);
	  
	  if (c < meshDistributor->getRstart() * nComponents || 
	      c >= meshDistributor->getRstart() * nComponents + nRankRows)
	    o_nnz[localRowIdx]++;
	  else
	    d_nnz[localRowIdx]++;
	}
      }
    }

    // The above algorithm for calculating the number of nnz per row over-
    // approximates the value, i.e., the number is always equal or larger to 
    // the real number of nnz values in the global parallel matrix. For small
    // matrices, the problem may arise, that the result is larger than the
    // number of elements in a row. This is fixed in the following.

    if (nRankRows < 100) 
      for (int i = 0; i < nRankRows; i++)
	d_nnz[i] = std::min(d_nnz[i], nRankRows);
  }


  void PetscSolver::fillPetscMatrix(Matrix<DOFMatrix*> *mat, SystemVector *vec)
  {
    FUNCNAME("PetscSolver::fillPetscMatrix()");

    double wtime = MPI::Wtime();
    int nRankRows = meshDistributor->getNumberRankDofs() * nComponents;
    int nOverallRows = meshDistributor->getNumberOverallDofs() * nComponents;

    // === Create PETSc vector (rhs, solution and a temporary vector). ===

    VecCreate(PETSC_COMM_WORLD, &petscRhsVec);
    VecSetSizes(petscRhsVec, nRankRows, nOverallRows);
    VecSetType(petscRhsVec, VECMPI);

    VecCreate(PETSC_COMM_WORLD, &petscSolVec);
    VecSetSizes(petscSolVec, nRankRows, nOverallRows);
    VecSetType(petscSolVec, VECMPI);

    VecCreate(PETSC_COMM_WORLD, &petscTmpVec);
    VecSetSizes(petscTmpVec, nRankRows, nOverallRows);
    VecSetType(petscTmpVec, VECMPI);

    int recvAllValues = 0;
    int sendValue = static_cast<int>(meshDistributor->getLastMeshChangeIndex() != lastMeshNnz);
    meshDistributor->getMpiComm().Allreduce(&sendValue, &recvAllValues, 1, MPI_INT, MPI_SUM);

    if (!d_nnz || recvAllValues != 0) {
      if (d_nnz) {
	delete [] d_nnz;
	d_nnz = NULL;
	delete [] o_nnz;
	o_nnz = NULL;
      }

      createPetscNnzStructure(mat);
      lastMeshNnz = meshDistributor->getLastMeshChangeIndex();
    }


    // === Create PETSc matrix with the computed nnz data structure. ===

    MatCreateMPIAIJ(PETSC_COMM_WORLD, nRankRows, nRankRows, nOverallRows, nOverallRows,
		    0, d_nnz, 0, o_nnz, &petscMatrix);
    
#if (DEBUG != 0)
    INFO(info, 8)("Fill petsc matrix 1 needed %.5f seconds\n", MPI::Wtime() - wtime);
#endif

#if (DEBUG != 0)
    int a, b;
    MatGetOwnershipRange(petscMatrix, &a, &b);
    TEST_EXIT(a == meshDistributor->getRstart() * nComponents)
      ("Wrong matrix ownership range!\n");
    TEST_EXIT(b == meshDistributor->getRstart() * nComponents + nRankRows)
      ("Wrong matrix ownership range!\n");
#endif


    // === Transfer values from DOF matrices to the PETSc matrix. === 

    for (int i = 0; i < nComponents; i++)
      for (int j = 0; j < nComponents; j++)
	if ((*mat)[i][j])
	  setDofMatrix((*mat)[i][j], nComponents, i, j);	
	
#if (DEBUG != 0)
    INFO(info, 8)("Fill petsc matrix 2 needed %.5f seconds\n", MPI::Wtime() - wtime);
#endif

    MatAssemblyBegin(petscMatrix, MAT_FINAL_ASSEMBLY);
    MatAssemblyEnd(petscMatrix, MAT_FINAL_ASSEMBLY);

#if (DEBUG != 0)
    INFO(info, 8)("Fill petsc matrix 3 needed %.5f seconds\n", 
		  TIME_USED(MPI::Wtime(), wtime));
#endif

    // === Transfer values from DOF vector to the PETSc vector. === 

    for (int i = 0; i < nComponents; i++)
      setDofVector(petscRhsVec, vec->getDOFVector(i), nComponents, i);

    VecAssemblyBegin(petscRhsVec);
    VecAssemblyEnd(petscRhsVec);

    INFO(info, 8)("Fill petsc matrix needed %.5f seconds\n", MPI::Wtime() - wtime);
  }


  void PetscSolver::solvePetscMatrix(SystemVector &vec, AdaptInfo *adaptInfo)
  {
    FUNCNAME("PetscSolver::solvePetscMatrix()");

#if 0
    // Set old solution to be initiual guess for petsc solver.
    for (int i = 0; i < nComponents; i++)
      setDofVector(petscSolVec, vec->getDOFVector(i), nComponents, i);

    VecAssemblyBegin(petscSolVec);
    VecAssemblyEnd(petscSolVec);
#endif

    // === Init Petsc solver. ===

    KSP solver;
    KSPCreate(PETSC_COMM_WORLD, &solver);
    KSPSetOperators(solver, petscMatrix, petscMatrix, SAME_NONZERO_PATTERN); 
    KSPSetTolerances(solver, 0.0, 1e-8, PETSC_DEFAULT, PETSC_DEFAULT);
    KSPSetType(solver, KSPBCGS);
    KSPMonitorSet(solver, myKSPMonitor, PETSC_NULL, 0);
    KSPSetFromOptions(solver);
    // Do not delete the solution vector, use it for the initial guess.
    //    KSPSetInitialGuessNonzero(solver, PETSC_TRUE);


    // === Run Petsc. ===

    KSPSolve(solver, petscRhsVec, petscSolVec);


    // === Transfere values from Petsc's solution vectors to the dof vectors.

    PetscScalar *vecPointer;
    VecGetArray(petscSolVec, &vecPointer);

    int nRankDofs = meshDistributor->getNumberRankDofs();
    for (int i = 0; i < nComponents; i++) {
      DOFVector<double> &dofvec = *(vec.getDOFVector(i));
      for (int j = 0; j < nRankDofs; j++)
	dofvec[meshDistributor->mapLocalToDofIndex(j)] = 
	  vecPointer[j * nComponents + i]; 
    }

    VecRestoreArray(petscSolVec, &vecPointer);


    // === Synchronize dofs at common dofs, i.e., dofs that correspond to more ===
    // === than one partition.                                                 ===
    meshDistributor->synchVector(vec);


    // === Print information about solution process. ===

    int iterations = 0;
    KSPGetIterationNumber(solver, &iterations);
    MSG("  Number of iterations: %d\n", iterations);
    adaptInfo->setSolverIterations(iterations);

    double norm = 0.0;
    MatMult(petscMatrix, petscSolVec, petscTmpVec);
    VecAXPY(petscTmpVec, -1.0, petscRhsVec);
    VecNorm(petscTmpVec, NORM_2, &norm);
    MSG("  Residual norm: %e\n", norm);


    // === Destroy Petsc's variables. ===

    MatDestroy(petscMatrix);
    VecDestroy(petscRhsVec);
    VecDestroy(petscSolVec);
    VecDestroy(petscTmpVec);
    KSPDestroy(solver);
  }

}
